package dev.fitko.fitconnect.core.events;

import com.nimbusds.jwt.SignedJWT;
import dev.fitko.fitconnect.api.config.ApplicationConfig;
import dev.fitko.fitconnect.api.domain.model.event.Event;
import dev.fitko.fitconnect.api.domain.model.event.EventLog;
import dev.fitko.fitconnect.api.domain.model.event.EventLogEntry;
import dev.fitko.fitconnect.api.domain.model.event.SubmissionStatus;
import dev.fitko.fitconnect.api.domain.model.event.authtags.AuthenticationTags;
import dev.fitko.fitconnect.api.domain.model.submission.Submission;
import dev.fitko.fitconnect.api.domain.validation.ValidationContext;
import dev.fitko.fitconnect.api.domain.validation.ValidationResult;
import dev.fitko.fitconnect.api.exceptions.internal.AuthenticationTagsEmptyException;
import dev.fitko.fitconnect.api.exceptions.internal.EventLogException;
import dev.fitko.fitconnect.api.exceptions.internal.RestApiException;
import dev.fitko.fitconnect.api.exceptions.internal.SubmitEventNotFoundException;
import dev.fitko.fitconnect.api.services.auth.OAuthService;
import dev.fitko.fitconnect.api.services.events.EventLogService;
import dev.fitko.fitconnect.api.services.events.EventLogVerificationService;
import dev.fitko.fitconnect.core.util.EventLogUtil;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.http.HttpEntity;
import org.springframework.http.HttpHeaders;
import org.springframework.http.HttpMethod;
import org.springframework.http.MediaType;
import org.springframework.web.client.RestClientException;
import org.springframework.web.client.RestTemplate;

import java.nio.charset.StandardCharsets;
import java.util.Comparator;
import java.util.List;
import java.util.UUID;
import java.util.stream.Collectors;

import static dev.fitko.fitconnect.core.util.EventLogUtil.eventEqualsSubmissionId;
import static dev.fitko.fitconnect.core.util.EventLogUtil.eventToLogEntry;
import static dev.fitko.fitconnect.core.util.EventLogUtil.getAuthenticationTags;
import static dev.fitko.fitconnect.core.util.EventLogUtil.getJWTSFromEvents;
import static dev.fitko.fitconnect.core.util.EventLogUtil.getJwtsFromEvent;
import static dev.fitko.fitconnect.core.util.EventLogUtil.mapEventLogToEntries;

public class EventLogApiService implements EventLogService {

    private static final Logger LOGGER = LoggerFactory.getLogger(EventLogApiService.class);
    public static final MediaType MEDIA_TYPE_JOSE = MediaType.parseMediaType("application/jose");

    private final ApplicationConfig config;
    private final OAuthService authService;
    private final RestTemplate restTemplate;
    private final EventLogVerificationService eventLogVerifier;

    public EventLogApiService(final ApplicationConfig config,
                              final OAuthService authService,
                              final RestTemplate restTemplate,
                              final EventLogVerificationService eventLogVerifier) {
        this.config = config;
        this.authService = authService;
        this.restTemplate = restTemplate;
        this.eventLogVerifier = eventLogVerifier;
    }

    @Override
    public List<EventLogEntry> getEventLog(final UUID caseId, final UUID destinationId) throws RestApiException, EventLogException {

        final EventLog eventLog = loadEventLog(caseId);
        final List<SignedJWT> jwtEvents = getJWTSFromEvents(eventLog.getEventLogs());
        final ValidationContext ctx = ValidationContext.withoutAuthTagValidation(destinationId, caseId);
        checkForValidationErrors(eventLogVerifier.validateEventLogs(ctx, jwtEvents));
        return mapEventLogToEntries(eventLog);
    }

    @Override
    public AuthenticationTags getAuthenticationTagsForEvent(final Event event, final Submission submission) throws RestApiException, EventLogException {

        final EventLog eventLog = loadEventLog(submission.getCaseId());
        final List<SignedJWT> submitEvents = getJwtsFromEvent(event, submission, eventLog);

        if (submitEvents.size() != 1) {
            throw new SubmitEventNotFoundException("Event log does not contain exactly one submit event");
        }

        final SignedJWT submitEvent = submitEvents.stream().findFirst().get();
        final AuthenticationTags authenticationTags = getAuthenticationTags(submitEvent);

        if (authenticationTags.getMetadata() == null && authenticationTags.getData() == null) {
            throw new AuthenticationTagsEmptyException("Authentication tags are empty");
        }

        final var ctx = ValidationContext.withAuthTagValidation(submission.getDestinationId(), submission.getCaseId(), authenticationTags);
        checkForValidationErrors(eventLogVerifier.validateEventLogs(ctx, List.of(submitEvent)));

        return authenticationTags;
    }

    @Override
    public SubmissionStatus getLastedEvent(final UUID destinationId, final UUID caseId, final UUID submissionId, final AuthenticationTags authenticationTags) throws EventLogException {

        final SignedJWT latestEvent = getLogFilteredBySubmission(caseId, submissionId);
        final ValidationContext contextWithAuthTagValidation = ValidationContext.withAuthTagValidation(destinationId, caseId, authenticationTags);
        return getVerifiedStatus(contextWithAuthTagValidation, latestEvent);
    }

    @Override
    public SubmissionStatus getLastedEventWithoutAuthTagValidation(final UUID destinationId, final UUID caseId, final UUID submissionId) throws EventLogException {

        final SignedJWT latestEvent = getLogFilteredBySubmission(caseId, submissionId);
        final ValidationContext contextWithoutAuthTagValidation = ValidationContext.withoutAuthTagValidation(destinationId, caseId);
        return getVerifiedStatus(contextWithoutAuthTagValidation, latestEvent);
    }

    @Override
    public void sendEvent(final UUID caseId, final String signedAndSerializedSET) {
        final String url = config.getEventsEndpoint();
        final HttpHeaders headers = getHttpHeaders(MEDIA_TYPE_JOSE);
        final HttpEntity<String> entity = new HttpEntity<>(signedAndSerializedSET, headers);
        try {
            restTemplate.exchange(url, HttpMethod.POST, entity, Void.class, caseId);
        } catch (final RestClientException e) {
            throw new RestApiException("Sending event failed", e);
        }
    }

    private EventLog loadEventLog(final UUID caseId) {
        final String url = config.getEventsEndpoint();
        final HttpHeaders headers = getHttpHeaders(MediaType.APPLICATION_JSON);
        final HttpEntity<String> entity = new HttpEntity<>(headers);
        try {
            return restTemplate.exchange(url, HttpMethod.GET, entity, EventLog.class, caseId).getBody();
        } catch (final RestClientException e) {
            throw new RestApiException("EventLog query failed", e);
        }
    }

    private HttpHeaders getHttpHeaders(final MediaType mediaType) {
        final var headers = new HttpHeaders();
        headers.setBearerAuth(authService.getCurrentToken().getAccessToken());
        headers.setContentType(mediaType);
        headers.setAcceptCharset(List.of(StandardCharsets.UTF_8));
        return headers;
    }

    private SubmissionStatus getVerifiedStatus(final ValidationContext validationContext, final SignedJWT latestEvent) {

        if (latestEvent == null) {
            LOGGER.info("No events found");
            return new SubmissionStatus();
        }
        checkForValidationErrors(eventLogVerifier.validateEventLogs(validationContext, List.of(latestEvent)));
        final EventLogEntry eventLogEntry = eventToLogEntry(latestEvent);
        return new SubmissionStatus(eventLogEntry.getEvent().getState(), eventLogEntry.getProblems());
    }

    private SignedJWT getLogFilteredBySubmission(final UUID caseId, final UUID submissionId) {
        final EventLog eventLog = loadEventLog(caseId);
        return getJWTSFromEvents(eventLog.getEventLogs()).stream()
                .filter(eventEqualsSubmissionId(submissionId))
                .max(Comparator.comparing(EventLogUtil::getEventFromJWT))
                .orElse(null);
    }

    private void checkForValidationErrors(final List<ValidationResult> validationResults) {
        if (!validationResults.isEmpty()) {
            throw new EventLogException(joinMessages(validationResults));
        }
    }

    private String joinMessages(final List<ValidationResult> validationResults) {
        return validationResults.stream()
                .map(ValidationResult::getError)
                .map(Exception::getMessage)
                .collect(Collectors.joining("\n"));
    }
}
