package dev.fitko.fitconnect.core.http;

import com.fasterxml.jackson.databind.ObjectMapper;
import dev.fitko.fitconnect.api.config.build.BuildInfo;
import dev.fitko.fitconnect.core.util.Strings;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.http.client.ClientHttpRequestInterceptor;
import org.springframework.http.client.SimpleClientHttpRequestFactory;
import org.springframework.http.converter.json.MappingJackson2HttpMessageConverter;
import org.springframework.web.client.RestTemplate;

import java.net.InetSocketAddress;
import java.net.Proxy;
import java.util.List;

import static com.fasterxml.jackson.databind.DeserializationFeature.FAIL_ON_UNKNOWN_PROPERTIES;

public class RestService {

    private static final Logger LOGGER = LoggerFactory.getLogger(RestService.class);

    private final BuildInfo buildInfo;
    private final String proxyHost;
    private final int proxyPort;

    public RestService(final String proxyHost, final int proxyPort, final BuildInfo buildInfo) {
        this.proxyHost = proxyHost;
        this.proxyPort = proxyPort;
        this.buildInfo = buildInfo;
    }

    public RestService(final BuildInfo buildInfo) {
        this(null, 0, buildInfo);
    }

    public RestTemplate getRestTemplate() {
        return hasProxySet() ? getProxyRestTemplate() : getDefaultRestTemplate();
    }

    private RestTemplate getDefaultRestTemplate() {
        LOGGER.info("No proxy configured");
        final RestTemplate restTemplate = new RestTemplate();
        setupTemplate(restTemplate);
        return restTemplate;
    }

    private RestTemplate getProxyRestTemplate() {
        LOGGER.info("Using proxy {}", this);
        final var requestFactory = new SimpleClientHttpRequestFactory();
        final var proxy = new Proxy(Proxy.Type.HTTP, new InetSocketAddress(proxyHost, proxyPort));
        requestFactory.setProxy(proxy);

        final RestTemplate proxyRestTemplate = new RestTemplate(requestFactory);
        setupTemplate(proxyRestTemplate);
        return proxyRestTemplate;
    }

    private void setMappingConverters(final RestTemplate restTemplate) {
        final MappingJackson2HttpMessageConverter jacksonMessageConverter = new MappingJackson2HttpMessageConverter();
        jacksonMessageConverter.setObjectMapper(new ObjectMapper().configure(FAIL_ON_UNKNOWN_PROPERTIES, false));
        restTemplate.getMessageConverters().add(jacksonMessageConverter);
    }

    private void setLoggingInterceptors(final RestTemplate restTemplate) {
        final List<ClientHttpRequestInterceptor> interceptors = restTemplate.getInterceptors();
        interceptors.add(new ApiRequestInterceptor());
        interceptors.add(new UserAgentInterceptor(buildInfo));
        restTemplate.setInterceptors(interceptors);
    }

    private void setupTemplate(final RestTemplate restTemplate) {
        setLoggingInterceptors(restTemplate);
        setMappingConverters(restTemplate);
    }

    boolean hasProxySet() {
        return !Strings.isNullOrEmpty(proxyHost) && proxyPort > 0;
    }

    @Override
    public String toString() {
        return String.format("ProxyConfig {host='%s', port=%d}", proxyHost, proxyPort);
    }
}
