package dev.fitko.fitconnect.core.submission;

import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.ObjectMapper;
import dev.fitko.fitconnect.api.config.ApplicationConfig;
import dev.fitko.fitconnect.api.domain.auth.OAuthToken;
import dev.fitko.fitconnect.api.domain.model.destination.Destination;
import dev.fitko.fitconnect.api.domain.model.submission.CreateSubmission;
import dev.fitko.fitconnect.api.domain.model.submission.Submission;
import dev.fitko.fitconnect.api.domain.model.submission.SubmissionForPickup;
import dev.fitko.fitconnect.api.domain.model.submission.SubmissionsForPickup;
import dev.fitko.fitconnect.api.domain.model.submission.SubmitSubmission;
import dev.fitko.fitconnect.api.exceptions.internal.RestApiException;
import dev.fitko.fitconnect.api.services.auth.OAuthService;
import dev.fitko.fitconnect.api.services.submission.SubmissionService;
import lombok.Builder;
import lombok.Getter;
import lombok.Setter;
import org.springframework.http.HttpEntity;
import org.springframework.http.HttpHeaders;
import org.springframework.http.HttpMethod;
import org.springframework.http.MediaType;
import org.springframework.web.client.RestClientException;
import org.springframework.web.client.RestTemplate;
import org.springframework.web.util.UriComponentsBuilder;

import java.nio.charset.StandardCharsets;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.UUID;

import static org.springframework.http.MediaType.APPLICATION_JSON_VALUE;

public class SubmissionApiService implements SubmissionService {

    private final OAuthService authService;
    private final RestTemplate restTemplate;
    private final ApplicationConfig config;

    public SubmissionApiService(final OAuthService authService, final RestTemplate restTemplate, final ApplicationConfig config) {
        this.authService = authService;
        this.restTemplate = restTemplate;
        this.config = config;
    }

    @Override
    public Destination getDestination(final UUID destinationID) throws RestApiException {
        final RequestSettings requestSettings = RequestSettings.builder()
                .url(config.getDestinationsEndpoint())
                .method(HttpMethod.GET)
                .responseType(Destination.class)
                .entity(getHttpEntity(getHeaders()))
                .params(Map.of("destinationId", destinationID))
                .build();
        return performRequest(requestSettings);
    }

    @Override
    public SubmissionForPickup announceSubmission(final CreateSubmission submission) throws RestApiException {
        final RequestSettings requestSettings = RequestSettings.builder()
                .url(config.getSubmissionsEndpoint())
                .method(HttpMethod.POST)
                .responseType(SubmissionForPickup.class)
                .entity(getHttpEntity(submission, getHeaders()))
                .params(Collections.emptyMap())
                .build();
        return performRequest(requestSettings);
    }

    @Override
    public void uploadAttachment(final UUID submissionId, final UUID attachmentId, final String encryptedAttachment) throws RestApiException {
        final Map<String, Object> params = new HashMap<>();
        params.put("submissionId", submissionId);
        params.put("attachmentId", attachmentId);
        final RequestSettings requestSettings = RequestSettings.builder()
                .url(config.getAttachmentEndpoint())
                .method(HttpMethod.PUT)
                .responseType(Void.class)
                .entity(getHttpEntity(encryptedAttachment, getHeaders("application/jose")))
                .params(params)
                .build();
        performRequest(requestSettings);
    }

    @Override
    public String getAttachment(final UUID submissionId, final UUID attachmentId) throws RestApiException {
        final Map<String, Object> params = new HashMap<>();
        params.put("submissionId", submissionId);
        params.put("attachmentId", attachmentId);
        final RequestSettings requestSettings = RequestSettings.builder()
                .url(config.getAttachmentEndpoint())
                .method(HttpMethod.GET)
                .responseType(String.class)
                .entity(getHttpEntity(getHeaders("application/jose")))
                .params(params)
                .build();
        return performRequest(requestSettings);
    }

    @Override
    public Submission sendSubmission(final SubmitSubmission submission) throws RestApiException {
        final Map<String, Object> params = Map.of("submissionId", submission.getSubmissionId());
        final RequestSettings requestSettings = RequestSettings.builder()
                .url(config.getSubmissionEndpoint())
                .method(HttpMethod.PUT)
                .responseType(Submission.class)
                .entity(getHttpEntity(submission, getHeaders()))
                .params(params)
                .build();
        return performRequest(requestSettings);
    }

    @Override
    public SubmissionsForPickup pollAvailableSubmissions(final int offset, final int limit) throws RestApiException {
        final String urlWithQueryParams = UriComponentsBuilder.fromHttpUrl(config.getSubmissionsEndpoint())
                .queryParam("limit", limit)
                .queryParam("offset", offset)
                .encode()
                .toUriString();
        final RequestSettings requestSettings = RequestSettings.builder()
                .url(urlWithQueryParams)
                .method(HttpMethod.GET)
                .responseType(SubmissionsForPickup.class)
                .entity(getHttpEntity(getHeaders()))
                .build();
        return performRequest(requestSettings);
    }

    @Override
    public SubmissionsForPickup pollAvailableSubmissionsForDestination(final UUID destinationId, final int offset, final int limit) {
        final String urlWithQueryParams = UriComponentsBuilder.fromHttpUrl(config.getSubmissionsEndpoint())
                .queryParam("destinationId", destinationId)
                .queryParam("limit", limit)
                .queryParam("offset", offset)
                .encode()
                .toUriString();
        final RequestSettings requestSettings = RequestSettings.builder()
                .url(urlWithQueryParams)
                .method(HttpMethod.GET)
                .responseType(SubmissionsForPickup.class)
                .entity(getHttpEntity(getHeaders()))
                .build();
        return performRequest(requestSettings);
    }

    @Override
    public Submission getSubmission(final UUID submissionId) throws RestApiException {
        final Map<String, Object> params = Map.of("submissionId", submissionId);
        final RequestSettings requestSettings = RequestSettings.builder()
                .url(config.getSubmissionEndpoint())
                .method(HttpMethod.GET)
                .responseType(Submission.class)
                .entity(getHttpEntity(getHeaders()))
                .params(params)
                .build();
        return performRequest(requestSettings);
    }

    private <T> T performRequest(final RequestSettings requestSettings) throws RestApiException {
        final String url = requestSettings.url;
        final HttpMethod method = requestSettings.method;
        final HttpEntity entity = requestSettings.entity;
        final Class<T> responseType = requestSettings.responseType;
        final Map<String, ?> params = requestSettings.params;
        try {
            return restTemplate.exchange(url, method, entity, responseType, params).getBody();
        } catch (final RestClientException e) {
            throw new RestApiException(e.getMessage(), e);
        }
    }

    private HttpEntity<String> getHttpEntity(final HttpHeaders headers) {
        return getHttpEntity(null, headers);
    }

    private HttpEntity<String> getHttpEntity(final Object body, final HttpHeaders headers) {
        if (body == null) {
            return new HttpEntity<>(headers);
        }
        try {
            return getHttpEntity(new ObjectMapper().writeValueAsString(body), headers);
        } catch (final JsonProcessingException e) {
            throw new RestApiException("Request body could not be processed", e);
        }
    }

    private HttpEntity<String> getHttpEntity(final String body, final HttpHeaders headers) {
        return body == null ? new HttpEntity<>(headers) : new HttpEntity<>(body, headers);
    }

    private HttpHeaders getHeaders() {
        return getHeaders(APPLICATION_JSON_VALUE);
    }

    private HttpHeaders getHeaders(final String contentType) {
        final OAuthToken token = authService.getCurrentToken();
        final var headers = new HttpHeaders();
        headers.setBearerAuth(token.getAccessToken());
        headers.setContentType(MediaType.parseMediaType(contentType));
        headers.setAcceptCharset(List.of(StandardCharsets.UTF_8));
        return headers;
    }

    @Getter
    @Setter
    @Builder
    private static class RequestSettings {
        private String url;
        private HttpMethod method;
        private HttpEntity entity;
        private Class responseType;
        @Builder.Default
        private Map<String, Object> params = new HashMap<>();
    }
}
